import warnings

import numpy
import onnx

from .utils import GLOBAL_VARS


def create_ndarray_f32(shape):
    return numpy.ones(shape, dtype=numpy.float32)


def create_ndarray_int64(shape):
    return numpy.zeros(shape, dtype=numpy.int64)


def shape_of_tensor(tensor):
    shape = []
    # for nb in tensor.shape.dim
    for nb in tensor.type.tensor_type.shape.dim:
        if nb.HasField('dim_value'):
            shape.append(nb.dim_value)
        if nb.HasField('dim_param'):
            shape.append(nb.dim_param)
    return shape


def shape_of_initializer(initial):
    shape = []
    # for nb in tensor.shape.dim
    for nb in initial.dims:
        shape.append(nb)
    return shape


def onnxdtype2npdtype(data_type):
    if data_type == onnx.TensorProto.FLOAT:
        return numpy.float32
    if data_type == onnx.TensorProto.DOUBLE:
        return numpy.float64
    if data_type == onnx.TensorProto.FLOAT16:
        return numpy.float16
    if data_type == onnx.TensorProto.INT32:
        return numpy.int32
    if data_type == onnx.TensorProto.INT16:
        return numpy.int16
    if data_type == onnx.TensorProto.INT64:
        return numpy.int64
    if data_type == onnx.TensorProto.INT8:
        return numpy.int8
    if data_type == onnx.TensorProto.UINT8:
        return numpy.uint8
    if data_type == onnx.TensorProto.BOOL:
        return numpy.bool_
    if data_type == onnx.TensorProto.STRING:
        return numpy.string_


def type_of_tensor(tensor):
    return onnxdtype2npdtype(tensor.type.tensor_type.elem_type)


def npdtype2onnxdtype(npdtype):
    if npdtype == numpy.float32:
        return onnx.TensorProto.FLOAT
    if npdtype == numpy.float64:
        return onnx.TensorProto.DOUBLE
    if npdtype == numpy.float16:
        return onnx.TensorProto.FLOAT16
    if npdtype == numpy.int32:
        return onnx.TensorProto.INT32
    if npdtype == numpy.int16:
        return onnx.TensorProto.INT16
    if npdtype == numpy.int64:
        return onnx.TensorProto.INT64
    if npdtype == numpy.int8:
        return onnx.TensorProto.INT8
    if npdtype == numpy.uint8:
        return onnx.TensorProto.UINT8
    if npdtype == numpy.bool_:
        return onnx.TensorProto.BOOL
    if npdtype == numpy.bytes_:
        return onnx.TensorProto.STRING
    if npdtype.type == numpy.string_:
        return onnx.TensorProto.STRING


def tensorproto2ndarray(initial):
    shape = shape_of_initializer(initial)
    ndtype = onnxdtype2npdtype(initial.data_type)
    if initial.raw_data == b'':
        arr = numpy.zeros(shape, ndtype).reshape((-1))
        if ndtype == numpy.float32:
            arr = numpy.fromiter(initial.float_data, dtype=ndtype)

        elif ndtype == numpy.int32:
            arr = numpy.fromiter(initial.int32_data, dtype=ndtype)

        elif ndtype == numpy.float16:
            raw = list(initial.int32_data)
            raw = numpy.fromiter(raw, dtype=numpy.uint16)
            mem = raw.tobytes()
            arr = numpy.frombuffer(mem, dtype=numpy.float16).reshape(shape)

        elif ndtype == numpy.int64:
            arr = numpy.fromiter(initial.int64_data, dtype=ndtype)

        elif ndtype == numpy.float64:
            arr = numpy.fromiter(initial.double_data, dtype=ndtype)

        elif ndtype == numpy.string_:
            arr = numpy.array(initial.string_data, dtype=ndtype)
    else:
        arr = numpy.frombuffer(initial.raw_data, dtype=ndtype)
    # if len(shape):
    arr = arr.reshape(shape)
    return arr


def get_attribute_data(att):
    if att.type == att.INTS:
        val = []
        for ints in att.ints:
            val.append(ints)
        return val
    elif att.type == att.INT:
        return att.i
    elif att.type == att.FLOAT:
        return att.f
    elif att.type == att.STRING:
        return att.s
    elif att.type == att.FLOATS:
        val = []
        for f in att.floats:
            val.append(f)
        return val
    elif att.type == att.TENSOR:
        return tensorproto2ndarray(att.t)


def volume(shape: []):
    # if not isinstance(shape,list):
    #     return 1 #scalar
    val = 1 if len(shape) > 0 else 0
    for v in shape:
        val *= v
    return val


def volume_tensor(t):
    if isinstance(t, numpy.ndarray):
        return volume(t.shape)
    return 1


def narray_calc_sparsity(arr):
    if len(arr.shape) != 2 and len(arr.shape) != 4:
        return 0
    if arr.dtype == numpy.float32 or arr.dtype == numpy.float64 or arr.dtype == numpy.int32 or arr.dtype == numpy.int8:
        flag = arr == 0
        return flag.sum() / arr.size
    if arr.dtype == numpy.uint8:
        flag = arr == 128
        return flag.sum() / arr.size
    if arr.dtype == numpy.float16:
        flag = arr == 0
        return flag.sum() / arr.size
    return 0


def narray_zero_flag(arr):
    if arr.dtype in (numpy.float32, numpy.float64, numpy.int32, numpy.int8, numpy.float16):
        flag = arr == 0
    if arr.dtype == numpy.uint8:
        flag = arr == 128
    return flag


def is_valid_ndarray(x):
    if x is None:
        return False
    if isinstance(x, (list, tuple)) and len(x) == 0:
        return False
    if isinstance(x, numpy.ndarray):
        if volume(x.shape) == 0:
            return True if x.size else False
        else:
            return True
    return False


def graph_addoutputs(graph: onnx.GraphProto, outnames: [str]) -> onnx.GraphProto:
    tensor_map = GLOBAL_VARS['tensor_map']
    for name in outnames:
        if tensor_map is not None and name in tensor_map.keys():
            newout = onnx.helper.make_tensor_value_info(name, onnx.TensorProto.FLOAT, tensor_map[name].shape)
        else:
            newout = onnx.helper.make_tensor_value_info(name, onnx.TensorProto.FLOAT, ())
        graph.output.append(newout)
    return graph


def graph_set_inputs(graph: onnx.GraphProto, dynamic_tensors: {}) -> onnx.GraphProto:
    tensor_map = GLOBAL_VARS['tensor_map']
    for input in graph.input:
        if dynamic_tensors.keys().__contains__(input.name):
            tensor_map[input.name] = dynamic_tensors[input.name]
            dim = input.type.tensor_type.shape.dim
            for nb, dnb in zip(dim, dynamic_tensors[input.name].shape):
                nb.dim_value = dnb
    return graph


def update_static_tensors(graph: onnx.GraphProto):
    tensor_map = GLOBAL_VARS['tensor_map']
    params_map = GLOBAL_VARS['params_map']
    for initial in graph.initializer:
        arr = tensorproto2ndarray(initial)
        tensor_map.update({initial.name: arr})

    for node in graph.node:
        if node.op_type == 'Constant':
            for att in node.attribute:
                if att.name == 'value':
                    tensor_map[node.output[0]] = get_attribute_data(att)

    totalparams = 0
    for key in tensor_map.keys():
        params_map[key] = volume(tensor_map[key].shape)
        totalparams += params_map[key]
    GLOBAL_VARS['totalparams'] = totalparams


def numpy_dtype2bytes(ndtype):
    return numpy.dtype(ndtype).itemsize


def search_sparse_blocksize(arr, ratio, deltar_thres=0.1):
    if len(arr.shape) == 2:  # gemm or matmul
        initsize = 2
        validsize = 1
        prevalid0 = True
        prevalid1 = True
        validratio = ratio
        while True:
            # try axis=1
            if prevalid1 and arr.shape[1] % initsize == 0:
                rearr = arr.reshape(arr.shape[0], -1, initsize)
                flag = narray_zero_flag(rearr)
                arrsum = numpy.sum(flag, -1)
                ratio1 = (arrsum == initsize).sum() / arrsum.size
                if ratio1 > ratio - deltar_thres:
                    valid1 = True
                    validratio = ratio1
                else:
                    valid1 = False
            else:
                valid1 = False

            # try axis=0
            if prevalid0 and arr.shape[0] % initsize == 0:
                rearr = arr.reshape(-1, initsize, arr.shape[1])
                flag = narray_zero_flag(rearr)
                arrsum = numpy.sum(flag, 1)
                ratio0 = (arrsum == initsize).sum() / arrsum.size
                if ratio0 > ratio - deltar_thres:
                    valid0 = True
                    validratio = ratio0
                else:
                    valid0 = False
            else:
                valid0 = False

            if not valid1 and not valid0:
                break
            validsize = initsize
            initsize *= 2
            prevalid0 = valid0
            prevalid1 = valid1

        # check square
        if prevalid1 and prevalid0:
            rearr = arr.reshape(arr.shape[0] // validsize, validsize, arr.shape[1] // validsize, validsize)
            flag = narray_zero_flag(rearr)
            arrsum = numpy.sum(flag, axis=(1, -1))
            ratios = (arrsum == (validsize * validsize)).sum() / arrsum.size
            if ratios > ratio - deltar_thres:
                return (validsize, validsize), ratios

        return (validsize if prevalid0 else 1, validsize if prevalid1 else 1), validratio

    if len(arr.shape) == 4:  # conv2d
        initsize = 2
        validsize = 1
        prevalid0 = True
        prevalid1 = True
        validratio0 = ratio
        validratio1 = ratio
        while True:
            # try axis=1
            if prevalid1 and arr.shape[1] % initsize == 0:
                rearr = arr.reshape(arr.shape[0], -1, initsize, *arr.shape[2:])
                flag = narray_zero_flag(rearr)
                arrsum = numpy.sum(flag, 2)
                ratio1 = (arrsum == initsize).sum() / arrsum.size
                if ratio1 > ratio - deltar_thres:
                    valid1 = True
                    validratio1 = ratio1
                else:
                    valid1 = False
            else:
                valid1 = False

            # try axis=0
            if prevalid0 and arr.shape[0] % initsize == 0:
                rearr = arr.reshape(-1, initsize, *arr.shape[1:])
                flag = narray_zero_flag(rearr)
                arrsum = numpy.sum(flag, 1)
                ratio0 = (arrsum == initsize).sum() / arrsum.size
                if ratio0 > ratio - deltar_thres:
                    valid0 = True
                    validratio0 = ratio0
                else:
                    valid0 = False
            else:
                valid0 = False

            if not valid1 and not valid0:
                break
            validsize = initsize
            initsize *= 2
            prevalid0 = valid0
            prevalid1 = valid1
        # check square
        if validsize > 1 and prevalid1 and prevalid0:
            rearr = arr.reshape(arr.shape[0] // validsize, validsize, arr.shape[1] // validsize, validsize,
                                *arr.shape[2:])
            flag = narray_zero_flag(rearr)
            arrsum = numpy.sum(flag, axis=(1, 3))
            ratios = (arrsum == (validsize * validsize)).sum() / arrsum.size
            if ratios > ratio - deltar_thres:
                return (validsize, validsize), ratios
        if validratio0 > validratio1:
            return (validsize, 1), validratio0
        return (1, validsize), validratio1

    return (1, 1), ratio


STATIC_TENSOR = 0
DYNAMIC_TENSOR = 1


class Tensor():

    def __init__(self, t):
        from .node import Node
        if isinstance(t, str):
            self.name = t
            self.proto = None
            self.shape = []
            self.numpy = None
            self.type = DYNAMIC_TENSOR
            self.dtype = numpy.float32
        elif isinstance(t, onnx.ValueInfoProto):
            self.name = t.name
            self.proto = t
            self.shape = shape_of_tensor(t)
            self.numpy = None
            self.type = DYNAMIC_TENSOR
            self.dtype = type_of_tensor(t)
        elif isinstance(t, onnx.TensorProto):
            self.name = t.name
            self.proto = t
            self.numpy = tensorproto2ndarray(t)
            self.shape = self.numpy.shape
            self.type = STATIC_TENSOR
            self.dtype = self.numpy.dtype.type
        else:
            assert 0
        self.sparsity_search()

    def update_tensor(self, data: numpy.ndarray):
        if not isinstance(data, numpy.ndarray):
            data = numpy.array(data)
        self.numpy = data
        self.update_shape(data.shape)
        self.update_dtype(self.numpy.dtype.type)

    def update_proto(self, data: numpy.ndarray):
        self.update_tensor(data)
        self.proto = self.make_tensor_proto()

    def update_shape(self, shape: list):
        if isinstance(shape, numpy.ndarray):
            assert 0
        self.shape = shape

    def update_dtype(self, dtype):
        self.dtype = dtype

    def shape2str(self):
        st = '['
        for val in self.shape:
            if isinstance(val, str):
                st += val + ','
            else:
                st += str(val) + ','
        st = st[:-1]
        st += ']'
        return st

    def get_shape(self):
        shape = []
        for s in self.shape:
            if isinstance(s, str):
                shape.append(s)
            else:
                shape.append(int(s))
        return shape

    def get_numpy(self):
        if self.numpy is not None:
            return self.numpy
        self.numpy = numpy.zeros(self.shape, dtype=self.dtype)
        return self.numpy

    def get_valueorshape(self):
        if self.numpy is not None:
            return self.numpy
        return self.shape

    def get_elementsize(self):
        if self.numpy is None or not isinstance(self.numpy, numpy.ndarray):
            return numpy_dtype2bytes(self.dtype)  # default as float
        return numpy_dtype2bytes(self.numpy.dtype)

    def get_memsize(self):
        return volume(self.get_shape()) * self.get_elementsize()

    def sparsity_search(self, thres_size=4096, thres_ratio=0.4):
        if self.type == DYNAMIC_TENSOR:
            self.sparsity = None
            return
        shape = self.get_shape()
        blocksize = (1, 1)
        blockratio = 0
        ratio = 0
        if (volume(shape) > thres_size) and self.numpy is not None:
            ratio = narray_calc_sparsity(self.numpy)
            if ratio is not None and ratio > thres_ratio:
                blocksize, blockratio = search_sparse_blocksize(self.numpy, ratio, deltar_thres=0.1)
        self.sparsity = {'blocksize': blocksize, 'blockratio': blockratio, 'ratio': ratio}

    def make_value_proto(self, make_dummy=False):
        shape = self.get_shape()
        if len(self.shape) == 0:
            shape = None
        if self.numpy is None:
            dtype = npdtype2onnxdtype(self.dtype)
        else:
            dtype = npdtype2onnxdtype(self.numpy.dtype)
        # shape = [int(i) for i in shape]
        vinf = onnx.helper.make_tensor_value_info(self.name, dtype, shape)
        return vinf

    def make_tensor_proto(self):
        if self.numpy is None:
            return None
        if len(self.numpy.shape) == 0:
            tproto = onnx.helper.make_tensor(self.name, npdtype2onnxdtype(self.numpy.dtype)
                                             , [], [self.numpy.item()])
        else:
            if self.numpy.dtype not in [numpy.float32, numpy.int32]:
                raw = True
                data = self.numpy.tobytes()
            else:
                raw = False
                data = self.numpy.flatten()
            tproto = onnx.helper.make_tensor(self.name, npdtype2onnxdtype(self.numpy.dtype)
                                             , self.numpy.shape, data, raw=raw)
        return tproto


def create_initial_Tensor(name: str, ndarray: numpy.ndarray):
    t = Tensor(name)
    t.type = STATIC_TENSOR
    t.numpy = ndarray
    t.shape = t.numpy.shape
    t.proto = t.make_tensor_proto()
    t.dtype = t.numpy.dtype.type
    return t



import os

import onnx

from .graph import Graph


class Model:
    def __init__(self, m: [str, onnx.ModelProto], verbose=False, constant_folding: bool = True,
                 noderename: bool = False):
        self.modelname = ''
        if isinstance(m, str):
            self.modelname = os.path.basename(m)
            self.modelname = os.path.splitext(self.modelname)[0]
            m = onnx.load_model(m)
        if not isinstance(m, onnx.ModelProto):
            self.valid = False
            return
        self.valid = True
        self.mproto = m
        self.graph = Graph(m.graph, verbose=verbose, constant_folding=constant_folding, noderename=noderename)

    def save_model(self, f: str, shape_only: bool = False, no_shape: bool = False):
        self.graph.save_model(f, shape_only=shape_only, rawmodel=self.mproto, no_shape=no_shape)



import argparse

import numpy

import onnx_tool


def get_parser():
    parser = argparse.ArgumentParser(
        "onnx_tool",
        description="Profile MACs(FLOPs) and perform shape inference on fixed-shape ONNX model"
    )
    parser.add_argument(
        "-m", "--mode",
        choices=['profile', 'export_tensors', 'constant_folding', 'io_modify'],
        default='profile',
        help="rm_iden: remove Identity layers")
    parser.add_argument(
        "-i", "--in", dest='in_', required=True,
        help="path of input ONNX model")
    parser.add_argument(
        "-o", "--out",
        help="path to save the ONNX model with shapes")
    parser.add_argument(
        '--names',
        nargs='+',
        default=None,
        help='tensor names as: --names 410 420'
    )
    parser.add_argument(
        '-d', '--dynamic_shapes',
        nargs='+',
        default=None,
        help='dynamic shape for io tensors as: --dynamic_shapes input:f32:1x3x224x224 scale:f32:1x2:0.25x0.25 '
             'or input:1x3x224x224 input:1x3xhxw'
    )
    parser.add_argument(
        "--fp16",
        action='store_true',
        help="path to save the ONNX model with shapes")
    parser.add_argument(
        "-f", "--file", default=None,
        help="file to store the MACs result for each node. None: print to console.")
    return parser


parser = get_parser()
args = parser.parse_args()


def __str2numpytype__(strtype):
    if strtype == 'f32':
        return numpy.float32
    if strtype == 'int32':
        return numpy.int32


def __str2list__(strshape, type):
    strs = strshape.split('x')
    shape = []
    for s in strs:
        shape.append(type(s))
    return shape


def __args2dynamicshapes__(args: [str]):
    dic = {}
    for arg in args:
        strs = arg.split(':')
        dtype = __str2numpytype__(strs[1])
        shape = __str2list__(strs[2], int)
        if len(strs) > 3:
            arr = numpy.array(__str2list__(strs[3], dtype), dtype=dtype).reshape(shape)
        else:
            arr = numpy.zeros(shape, dtype=dtype)
        dic[strs[0]] = arr
    return dic


def __args2strshapes__(args: [str]):
    dic = {}
    for arg in args:
        strs = arg.split(':')
        dic[strs[0]] = strs[1]
    return dic


if args.mode == 'profile':
    if args.dynamic_shapes is not None:
        dynamic = __args2dynamicshapes__(args.dynamic_shapes)
    else:
        dynamic = None
    onnx_tool.model_profile(args.in_, dynamic, args.file, args.out)
elif args.mode == 'export_tensors':
    onnx_tool.model_export_tensors_numpy(args.in_, tensornames=args.names, savefolder=args.out, fp16=args.fp16)
elif args.mode == 'constant_folding':
    onnx_tool.model_constant_folding(args.in_, args.out)
elif args.mode == 'io_modify':
    if args.dynamic_shapes is not None:
        shapedic = __args2strshapes__(args.dynamic_shapes)
        onnx_tool.model_io_modify(args.in_, args.out, shapedic)



import os
import warnings

import numpy
import onnx

from .graph import Graph
from .model import Model
from .node import NODE_REGISTRY, Node
from .serialization import serialize_shape_engine, serialize_graph
from .tensor import create_ndarray_f32, create_ndarray_int64
from .utils import timer, VERSION


def loadmodel(f, verbose: bool = False, constant_folding: bool = True, noderename: bool = False):
    model = Model(f, verbose=verbose, constant_folding=constant_folding, noderename=noderename)
    if model.valid is False:
        warnings.warn(f'Invalid onnx model file:{f}')
        exit(-1)
    return model


def __remove_initilisers(graph: onnx.GraphProto):
    graph.ClearField('initializer')


def model_export_tensors_numpy(m, tensornames: [str] = None, savefolder: str = None, fp16: bool = False) -> None:
    model = loadmodel(m)
    def save_numpy(arr: numpy.ndarray, fp16: bool, filename):
        if fp16 and arr.dtype in [numpy.float32, numpy.float64]:
            arr = arr.astype(numpy.float16)
        numpy.save(filename, arr)

    g = model.graph
    if savefolder is not None:
        os.makedirs(savefolder, exist_ok=True)
    else:
        savefolder = './'
    if tensornames is None:
        for key in g.initials:
            name = key
            if '/' in key:
                name = key.replace('/', '_')
            if '\\' in key:
                name = key.replace('\\', '_')
            narr = g.tensormap[key].numpy
            save_numpy(narr, fp16, os.path.join(savefolder, name + '.npy'))

    else:
        for name in tensornames:
            if name not in g.initials:
                warnings.warn(f'tensor {name} not found ')
                continue
            fname = name
            if '/' in name:
                fname = name.replace('/', '_')
            if '\\' in name:
                fname = name.replace('\\', '_')
            narr = g.tensormap[name].numpy
            save_numpy(narr, fp16, os.path.join(savefolder, fname + '.npy'))


# These ops are created by onnx exporter, they are out of programmer's sense
DefaultFilter = (
    'Identity', 'Constant',
)

# These ops have no computation
NoMacsOps = (
    'Identity', 'Constant', 'Shape', 'Squeeze', 'Unsqueeze', 'Reshape', 'ConstantOfShape', 'Cast', 'Pad', 'Concat',
    'Slice', 'Gather'
)


def model_profile(m, dynamic_shapes: {str: tuple} = None, savenode: str = None,
                  saveshapesmodel: str = None, shapesonly: bool = False, verbose: bool = False,
                  constant_folding: bool = False,
                  hidden_ops: [str] = NoMacsOps) -> None:
    model = loadmodel(m, verbose=verbose, constant_folding=constant_folding)
    g = model.graph
    gtmr = timer()
    g.graph_reorder_nodes()
    gtmr.start()
    g.shape_infer(dynamic_shapes)
    g.log(f'infered all tensor shapes, time cost {gtmr.stop():.3f} s')
    gtmr.start()
    g.profile()
    g.log(f'profile all nodes, time cost {gtmr.stop():.3f} s')
    g.print_node_map(savenode, exclude_ops=hidden_ops)
    if saveshapesmodel is not None:
        model.save_model(saveshapesmodel, shape_only=shapesonly)


def model_shape_regress(m, input_desc: {}, input_range: {}):
    model = loadmodel(m)
    graph = model.graph
    graph.graph_reorder_nodes()
    shape_engine = graph.shape_regress(input_desc, input_range)
    cg = graph.get_compute_graph()
    return shape_engine, cg

def model_constant_folding(m, f: str):
    model = loadmodel(m,constant_folding=True,verbose=True)
    model.save_model(f)


def model_shape_infer(m, dynamic_shapes: {str: tuple} = None,
                      saveshapesmodel: str = None, shapesonly: bool = False, no_shape: bool = False,
                      verbose: bool = False):
    model = loadmodel(m,verbose=verbose)
    model.graph.shape_infer(dynamic_shapes)
    if saveshapesmodel is not None:
        model.save_model(saveshapesmodel, shape_only=shapesonly, no_shape=no_shape)


def model_reorder_nodes(m, savemodel: str, ):
    model = loadmodel(m)
    model.graph.graph_reorder_nodes()
    model.save_model(savemodel)


def model_io_modify(m, savemodel: str, custom_io):
    '''
        Args:
            m: onnx.ModelProto or file path
            custom_io: {str:str} e.g. {'input':'Nx3xwidthxheight'}
        Returns:

    '''
    model = loadmodel(m)
    graph = model.mproto.graph
    if custom_io is not None:
        keylist = list(custom_io.keys())
        for i, input in enumerate(graph.input):
            if input.name in keylist:
                shapes = custom_io[input.name].split('x')
                # maybe consider create a new valueinfoproto
                dim = input.type.tensor_type.shape.dim
                assert (len(shapes) == len(dim))
                for nb, shapeval in zip(dim, shapes):
                    if shapeval.isnumeric():
                        if nb.HasField('dim_param'):
                            nb.ClearField('dim_param')
                        nb.dim_value = int(shapeval)
                    else:
                        if nb.HasField('dim_value'):
                            nb.ClearField('dim_value')
                        nb.dim_param = shapeval

        for i, output in enumerate(graph.output):
            if output.name in keylist:
                shapes = custom_io[output.name].split('x')
                # maybe consider create a new valueinfoproto
                dim = output.type.tensor_type.shape.dim
                assert (len(shapes) == len(dim))
                for nb, shapeval in zip(dim, shapes):
                    if shapeval.isnumeric():
                        if nb.HasField('dim_param'):
                            nb.ClearField('dim_param')
                        nb.dim_value = int(shapeval)
                    else:
                        if nb.HasField('dim_value'):
                            nb.ClearField('dim_value')
                        nb.dim_param = shapeval
    graph = Graph(graph)
    graph.save_model(savemodel, rawmodel=m)


def model_subgraph(m, in_tensor_names: [str] = None, out_tensor_names: [str] = None, nodenames: [str] = None,
                   savefolder='./'):
    model = loadmodel(m)
    graph = model.graph
    if in_tensor_names is not None and out_tensor_names is not None:
        graph_lvl0, graph_lvl1, graph_lvl2 = graph.get_subgraph(inputs=in_tensor_names, outputs=out_tensor_names)
        graph_lvl0.save_model(os.path.join(savefolder, model.modelname + '_level0.onnx'), rawmodel=model.mproto)
        graph_lvl1.save_model(os.path.join(savefolder, model.modelname + '_level1.onnx'), rawmodel=model.mproto)
        graph_lvl2.save_model(os.path.join(savefolder, model.modelname + '_level2.onnx'), rawmodel=model.mproto)
    if nodenames is not None:
        rawgraph = graph.get_onnxgraph_by_nodenames(nodenames)
        subgraph = Graph(rawgraph)
        subgraph.save_model(os.path.join(savefolder, model.modelname + '_subgraph.onnx'), rawmodel=model.mproto)


def model_opfusion(m, op_type: str, op_name: str, savefile, in_tensor_names: [str] = None,
                   out_tensor_names: [str] = None, nodenames: [str] = None, keep_attr=True):
    model = loadmodel(m)
    graph = model.graph
    if in_tensor_names is not None and out_tensor_names is not None:
        graph = graph.fuse_subgraph_iotensors(inputs=in_tensor_names, outputs=out_tensor_names, name=op_name,
                                              nodeop=op_type, keep_attr=keep_attr)
        model.save_model(savefile)
    if nodenames is not None:
        graph = graph.fuse_subgraph_node_names(nodenames, nodeop=op_type, name=op_name, keep_attr=keep_attr)
        model.save_model(savefile)



#!/usr/bin/env python

from setuptools import setup, find_packages

readme = open("README.md", encoding="utf-8").read()
VERSION = "0.8.1"

requirements = [
    "onnx",
    "numpy",
    'tabulate'
]

setup(
    # Metadata
    name="onnx-tool",
    version=VERSION,
    author="Luo Yu",
    author_email="luoyu888888@gmail.com",
    url="https://github.com/ThanatosShinji/onnx-tool",
    description="A tool for ONNX model: A parser, editor and profiler tool for ONNX models.",
    long_description=readme,
    long_description_content_type="text/markdown",
    license="MIT",
    # Package info
    packages=find_packages(),
    #
    zip_safe=True,
    install_requires=requirements,
    # Classifiers
    classifiers=[
        "Programming Language :: Python :: 3",
    ],
)



import numpy
import onnx.helper

import onnx_tool
from .graph import Graph
from .node import Node

ConvBN = [
    {
        'name': 'conv_0',
        'op': 'Conv',
        'attrs': [
            ['kernel_shape', '<=', 0, 7],
            ['kernel_shape', '>=', 0, 1],
            ['kernel_shape', '<=', 1, 7],
            ['kernel_shape', '>=', 1, 1],
        ],
        'inport': [],
        'outport': [[0, 'bn_0', 0]],
    },
    {
        'name': 'bn_0',
        'op': 'BatchNormalization',
        'attrs': [
        ],
        'inport': [[0, 'conv_0', 0]],
        'outport': [],
    },
]

Fused_Element = [
    {
        'name': 'any',
        'op': 'Any',
        'attrs': [],
        'inport': [],
        'outport': [[0, 'act_0', 0]],
    },
    {
        'name': 'act_0',
        'op': ['Relu', 'LeakyRelu', 'Add'],
        'attrs': [
        ],
        'inport': [[0, 'any', 0]],
        'outport': [],
    },
]

Conv_Res = [
    {
        'name': 'conv_0',
        'op': 'Conv',
        'attrs': [],
        'inport': [],
        'outport': [[0, 'add_0', 1]],
    },
    {
        'name': 'add_0',
        'op': ['Add', ],
        'attrs': [],
        'inport': [[1, 'conv_0', 0]],
        'outport': [[0, 'act_0', 0]],
    },
    {
        'name': 'act_0',
        'op': ['Relu', 'LeakyRelu'],
        'attrs': [],
        'inport': [[0, 'add_0', 0]],
        'outport': [],
    },
]

MHAint8_Pattern = [
    {
        'name': 'MatMul_0',
        'op': 'QLinearMatMul',
        'attrs': [
        ],
        'inport': [],
        'outport': [[0, 'DequantizeLinea_0', 0]],
    },
    {
        'name': 'DequantizeLinea_0',
        'op': 'DequantizeLinear',
        'attrs': [
        ],
        'inport': [[0, 'MatMul_0', 0]],
        'outport': [[0, 'div_0', 0]],
    },
    {
        'name': 'div_0',
        'op': 'Div',
        'attrs': [
        ],
        'inport': [[0, 'DequantizeLinea_0', 0]],
        'outport': [[0, 'Add_0', 0]],
    },
    {
        'name': 'Add_0',
        'op': 'Add',
        'attrs': [
        ],
        'inport': [[0, 'div_0', 0]],
        'outport': [[0, 'Softmax_0', 0]],
    },
    {
        'name': 'Softmax_0',
        'op': 'Softmax',
        'attrs': [
        ],
        'inport': [[0, 'Add_0', 0]],
        'outport': [[0, 'QuantizeLinear_1', 0]],
    },
    {
        'name': 'QuantizeLinear_1',
        'op': 'QuantizeLinear',
        'attrs': [
        ],
        'inport': [[0, 'Add_0', 0]],
        'outport': [[0, 'MatMul_1', 0]],
    },
    {
        'name': 'MatMul_1',
        'op': 'QLinearMatMul',
        'attrs': [
        ],
        'inport': [[0, 'QuantizeLinear_1', 0]],
        'outport': [],
    },
]

layernorm_pattern = [
    {
        'name': 'ReduceMean_196',
        'op': 'ReduceMean',
        'attrs': [],
        'inport': [],
        'outport': [[0, 'Sub_197', 1]]
    },
    {
        'name': 'Sub_197',
        'op': 'Sub',
        'attrs': [],
        'inport': [[1, 'ReduceMean_196', 0]],
        'outport': [[0, 'Pow_0', 0], [0, 'Div_0', 0]]
    },
    {
        'name': 'Pow_0',
        'op': 'Pow',
        'attrs': [],
        'inport': [[0, 'Sub_197', 0]],
        'outport': [[0, 'ReduceMean_0', 0]]
    },
    {
        'name': 'ReduceMean_0',
        'op': 'ReduceMean',
        'attrs': [],
        'inport': [[0, 'Pow_0', 0]],
        'outport': [[0, 'Add_0', 0]]
    },
    {
        'name': 'Add_0',
        'op': 'Add',
        'attrs': [],
        'inport': [[0, 'ReduceMean_0', 0]],
        'outport': [[0, 'Sqrt_0', 0]]
    },
    {
        'name': 'Sqrt_0',
        'op': 'Sqrt',
        'attrs': [],
        'inport': [[0, 'Add_0', 0]],
        'outport': [[0, 'Div_0', 1]]
    },
    {
        'name': 'Div_0',
        'op': 'Div',
        'attrs': [],
        'inport': [[0, 'Sub_197', 0], [1, 'Sqrt_0', 0]],
        'outport': []
    },
]

ShapeOps = ['Flatten', 'Reshape', 'Unsqueeze', 'Squeeze']


def removeShapeOps(g: onnx_tool.Graph):
    rmlist = []
    for n in g.nodemap.keys():
        node = g.nodemap[n]
        if node.op_type in ShapeOps:
            rmlist.append(n)
    for n in rmlist:
        g.skip_node(n)
        g.nodemap.pop(n)
    g.update_tensor_relations()
    return g


def createSerialOpChain(oplist: list[str]):
    chain = []
    for i, op in enumerate(oplist):
        inport = [] if i == 0 else [[-1, str(i - 1), -1]]
        outport = [] if i == len(oplist) - 1 else [[-1, str(i + 1), -1]]
        nodedesc = {
            'name': str(i),
            'op': op,
            'attrs': [],
            'inport': inport,
            'outport': outport
        }
        chain.append(nodedesc)
    return chain


def createSerialPattern(oplist: list[str]):
    chain = createSerialOpChain(oplist)
    return FusionPattern(chain)


class AttrExpr():
    def __init__(self, raw: []):
        self.attrname = raw[0]
        self.expr = raw[1]
        if len(raw) == 4:
            self.idx = raw[2]
            self.num = raw[3]
        else:
            self.num = raw[2]
            self.idx = -1

    def __call__(self, x):
        if hasattr(x, self.attrname):
            if self.idx == -1:
                return self.logical(x.__getattribute__(self.attrname))
            else:
                return self.logical(x.__getattribute__(self.attrname)[self.idx])
        return False

    def logical(self, attr):
        if self.expr == '<=':
            return attr <= self.num
        if self.expr == '<':
            return attr < self.num
        if self.expr == '>=':
            return attr >= self.num
        if self.expr == '>':
            return attr > self.num
        if self.expr == '==':
            return attr == self.num


class NodeCondition():
    def __init__(self, attrmap: {}):
        self.op = attrmap['op']
        self.name = attrmap['name']
        self.attexprs = []
        for att in attrmap['attrs']:
            self.attexprs.append(AttrExpr(att))
        self.outport = attrmap['outport']
        self.inport = attrmap['inport']

    def is_node(self, node: Node):
        flag = True

        if isinstance(self.op, list):
            flag &= node.op_type in self.op
        else:
            if self.op != 'Any':
                flag &= node.op_type == self.op
        for attrexpr in self.attexprs:
            flag &= attrexpr(node)
        return flag

def create_descs_from_nodenames(graph:Graph, nodenames:[str]):
    nodedesc=[]
    consumed_by={}
    produced_by={}
    for name in nodenames:
        node = graph.nodemap[name]
        desc ={
            'name':node.name,
            'op':node.op_type,
            'attrs': [],
            'inport': [],
            'outport': [],
        }
        nodedesc.append(desc)
        for t in node.output:
            consumed_by[t]=[]
            produced_by[t]=name

    for i,name in enumerate(nodenames):
        node = graph.nodemap[name]
        for j,t in enumerate(node.input):
            if t in consumed_by.keys():
                pnode=produced_by[t]
                pidx=nodenames.index(pnode)
                prodnode=graph.nodemap[pnode]
                nodedesc[pidx]['outport'].append([prodnode.output.index(t),name,j])
                nodedesc[i]['inport'].append([j,pnode,prodnode.output.index(t)])
    return nodedesc

class FusionPattern():
    def __init__(self, nodedescs: {}, inplace_fusion=False):
        self.nodedesc = {}
        self.first_key = nodedescs[0]['name']
        self.inplace_fusion = inplace_fusion
        self.append_fusion = inplace_fusion
        for desc in nodedescs:
            self.nodedesc[desc['name']] = NodeCondition(desc)

    def search_node(self, nodepair, graph, searched):
        curdescname=nodepair[0]
        curnodename=nodepair[1]

        desc = self.nodedesc[curdescname]
        node = graph.nodemap[curnodename]
        searched.append(curnodename)

        # expand in nodes
        invalid=False if len(desc.inport)>0 else True
        uppaths=[]
        for inset in desc.inport:
            inidx = inset[0]
            indesckey = inset[1]
            prev_outidx = inset[2]
            nextdesc = self.nodedesc[indesckey]
            if inidx == -1:
                for tname in node.input:
                    if tname not in graph.producedby:
                        continue
                    producer_node = graph.producedby[tname]
                    for nodename in producer_node:
                        nodeobject = graph.nodemap[nodename]
                        if prev_outidx == -1 or nodeobject.output.index(tname) == prev_outidx:
                            if nextdesc.is_node(nodeobject):
                                if nodename not in searched:
                                    invalid, uppath = self.search_node((indesckey, nodename), graph, searched)
                                    if invalid:
                                        uppaths.append(uppath)
                                        break
                                else:
                                    invalid = True
            elif inidx < len(node.input):
                tname = node.input[inidx]
                producer_node = graph.producedby[tname]
                for nodename in producer_node:
                    nodeobject = graph.nodemap[nodename]
                    if prev_outidx == -1 or nodeobject.output.index(tname) == prev_outidx:
                        if nextdesc.is_node(nodeobject):
                            if nodename not in searched:
                                invalid, uppath = self.search_node((indesckey, nodename), graph, searched)
                                if invalid:
                                    uppaths.append(uppath)
                                    break
                            else:
                                invalid = True

        if not invalid:
            searched.remove(curnodename)
            return False, None
        outpath = []
        for uppath in uppaths:
            if uppath is not None:
                for v in uppath:
                    outpath.append(v)
        outpath.append(nodepair)

        outvalid=False if len(desc.outport)>0 else True
        downpaths=[]
        for outset in desc.outport:
            outidx = outset[0]
            outdescky = outset[1]
            next_inidx = outset[2]
            nextdesc = self.nodedesc[outdescky]
            if outidx == -1:
                for output in node.output:
                    if outvalid:
                        break
                    if output in graph.consumedby:
                        consumed_nodes = graph.consumedby[output]
                        if self.append_fusion and len(consumed_nodes) > 1:
                            # inpalce_fusion the consumer op will be appended to this op as postop
                            # it requires that the output of this op is consumed by next op only
                            continue
                        for nodename in consumed_nodes:
                            nodeobject = graph.nodemap[nodename]
                            if next_inidx == -1 or nodeobject.input.index(output) == next_inidx:
                                if nextdesc.is_node(nodeobject):
                                    if nodename not in searched:
                                        outvalid,downpath=self.search_node((outdescky,nodename),graph,searched)
                                        if outvalid:
                                            downpaths.append(downpath)
                                            break
                                    else:
                                        outvalid=True


            elif outidx < len(node.output):
                tname = node.output[outidx]
                if tname in graph.consumedby:
                    consumed_nodes = graph.consumedby[tname]
                    if self.append_fusion and len(consumed_nodes) > 1:
                        # inpalce_fusion the consumer op will be appended to this op as postop
                        # it requires that the output of this op is consumed by next op only
                        continue
                    for nodename in consumed_nodes:
                        nodeobject = graph.nodemap[nodename]
                        if next_inidx == -1 or nodeobject.input.index(tname) == next_inidx:
                            if nextdesc.is_node(nodeobject):
                                if nodename not in searched:
                                    outvalid, downpath = self.search_node((outdescky, nodename), graph, searched)
                                    if outvalid:
                                        downpaths.append(downpath)
                                        break
                                else:
                                    outvalid = True
                                    break

        if outvalid:
            for downpath in downpaths:
                if downpath is not None:
                    for v in downpath:
                        outpath.append(v)
            return True, outpath
        else:
            searched.remove(curnodename)
            return False, None

    def search_pattern(self, graph: Graph):
        ls_nodes = []
        first_desc = self.nodedesc[self.first_key]
        self.found_node_names = []
        self.tmp_names = []
        for name in graph.nodemap.keys():
            if name in self.found_node_names:
                continue
            node = graph.nodemap[name]
            if first_desc.is_node(node):
                searched=[]
                valid, path = self.search_node((self.first_key, name), graph, searched)
                if valid:
                    desckeys=list(self.nodedesc.keys())
                    nodes=['a']*len(desckeys)
                    for val in path:
                        idx = desckeys.index(val[0])
                        nodes[idx]=val[1]
                    ls_nodes.append(nodes)
        return ls_nodes






def ConvBNFusion(graph: Graph):
    import math
    pattern = FusionPattern(ConvBN)
    node_sets = pattern.search_pattern(graph)
    for nodes in node_sets:
        conv_node = graph.nodemap[nodes[0]]
        bn_node = graph.nodemap[nodes[1]]
        weight = graph.tensormap[conv_node.input[1]]
        gamma = graph.tensormap[bn_node.input[1]]
        beta = graph.tensormap[bn_node.input[2]]
        mean = graph.tensormap[bn_node.input[3]]
        var = graph.tensormap[bn_node.input[4]]
        wshape = weight.shape
        if len(conv_node.input) == 3:
            hasbias = True
            newbias = graph.nodemap[conv_node.input[2]].numpy
        else:
            hasbias = False
            newbias = numpy.zeros((wshape[0],), dtype=numpy.float32)
        for i in range(wshape[0]):
            sm = gamma.numpy[i] / math.sqrt(var.numpy[i] + bn_node.epsilon)
            sv = beta.numpy[i]
            for j in range(wshape[1]):
                for k in range(wshape[2]):
                    for l in range(wshape[3]):
                        weight.numpy[i, j, k, l] = sm * weight.numpy[i, j, k, l]
            newbias[i] = sm * (newbias[i] - mean.numpy[i]) + sv
        weight.update_proto(weight.numpy)
        newbiasname = conv_node.name + "_bias_"
        graph.add_initial(newbiasname, newbias)
        graph.remove_node(nodes[1])
        if hasbias:
            conv_node.input[2] = newbiasname
        else:
            conv_node.input.append(newbiasname)
        graph.producedby[bn_node.output[0]] = [conv_node.name]
        conv_node.output[0] = bn_node.output[0]
